/*
 * Copyright (C) 2016 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.akamai.downloader.sample;

import android.app.Activity;
import android.app.Fragment;
import android.content.Context;
import android.content.Intent;
import android.content.res.AssetManager;
import android.net.Uri;
import android.os.AsyncTask;
import android.os.Bundle;
import android.os.Handler;
import android.support.annotation.Nullable;
import android.util.JsonReader;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.ViewGroup;
import android.widget.BaseExpandableListAdapter;
import android.widget.ExpandableListView;
import android.widget.ExpandableListView.OnChildClickListener;
import android.widget.FrameLayout;
import android.widget.ImageButton;
import android.widget.TextView;
import android.widget.Toast;

import com.akamai.amp.downloader.StreamTracksListener;
import com.akamai.amp.downloader.exo.AmpDownloadService;
import com.akamai.amp.downloader.idea.AmpDownload;
import com.akamai.amp.downloader.idea.AmpDownloadManager;
import com.akamai.amp.downloader.idea.AmpDownloaderListener;
import com.akamai.amp.media.exowrapper2.ExoWrapper;
import com.akamai.exoplayer2.ParserException;
import com.akamai.exoplayer2.offline.DownloadService;
import com.akamai.exoplayer2.offline.TrackKey;
import com.akamai.exoplayer2.source.smoothstreaming.manifest.SsUtil;
import com.akamai.exoplayer2.upstream.DataSource;
import com.akamai.exoplayer2.upstream.DataSourceInputStream;
import com.akamai.exoplayer2.upstream.DataSpec;
import com.akamai.exoplayer2.upstream.DefaultDataSource;
import com.akamai.exoplayer2.util.Assertions;
import com.akamai.exoplayer2.util.Log;
import com.akamai.exoplayer2.util.Util;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/** An activity for selecting from a list of media samples. */
public class SampleChooserActivity extends Activity
        implements OnChildClickListener {

  private static final String TAG = "SampleChooserActivity";

  private boolean useExtensionRenderers;
  private SampleAdapter sampleAdapter;
  private MenuItem preferExtensionDecodersMenuItem;
  private MenuItem randomAbrMenuItem;
  private FrameLayout loadingFragment;

  @Override
  public void onCreate(Bundle savedInstanceState) {
    super.onCreate(savedInstanceState);
    setContentView(R.layout.sample_chooser_activity);
    sampleAdapter = new SampleAdapter();
    ExpandableListView sampleListView = findViewById(R.id.sample_list);
    loadingFragment = findViewById(R.id.loading_container);
    sampleListView.setAdapter(sampleAdapter);
    sampleListView.setOnChildClickListener(this);

    Intent intent = getIntent();
    String dataUri = intent.getDataString();
    String[] uris;
    if (dataUri != null) {
      uris = new String[] {dataUri};
    } else {
      ArrayList<String> uriList = new ArrayList<>();
      AssetManager assetManager = getAssets();
      try {
        for (String asset : assetManager.list("")) {
          if (asset.endsWith(".exolist.json")) {
            uriList.add("asset:///" + asset);
          }
        }
      } catch (IOException e) {
        Toast.makeText(getApplicationContext(), R.string.sample_list_load_error, Toast.LENGTH_LONG)
                .show();
      }
      uris = new String[uriList.size()];
      uriList.toArray(uris);
      Arrays.sort(uris);
    }

    SampleApplication application = (SampleApplication) getApplication();
    //useExtensionRenderers = application.useExtensionRenderers();
    useExtensionRenderers = false;
    SampleListLoader loaderTask = new SampleListLoader();
    loaderTask.execute(uris);

    // Start the download service if it should be running but it's not currently.
    // Starting the service in the foreground causes notification flicker if there is no scheduled
    // action. Starting it in the background throws an exception if the app is in the background too
    // (e.g. if device screen is locked).
    try {
      DownloadService.start(this, AmpDownloadService.class);
    } catch (IllegalStateException e) {
      DownloadService.startForeground(this, AmpDownloadService.class);
    }
    setupLoading();
    loadingFragment.setVisibility(View.INVISIBLE);
  }

  private void setupLoading() {
    Bundle b = new Bundle();
    b.putCharSequence(LoadingFragment.DESCRIPTION_KEY, "Loading Tracks");

    Fragment f = new LoadingFragment();
    f.setArguments(b);

    getFragmentManager()
            .beginTransaction()
            .replace(R.id.loading_container, f)
            .commit();
  }

  @Override
  public boolean onCreateOptionsMenu(Menu menu) {
    MenuInflater inflater = getMenuInflater();
    inflater.inflate(R.menu.sample_chooser_menu, menu);
    preferExtensionDecodersMenuItem = menu.findItem(R.id.prefer_extension_decoders);
    preferExtensionDecodersMenuItem.setVisible(useExtensionRenderers);
    randomAbrMenuItem = menu.findItem(R.id.random_abr);
    return true;
  }

  @Override
  public boolean onOptionsItemSelected(MenuItem item) {
    item.setChecked(!item.isChecked());
    if(item.getItemId() == R.id.random_abr){
      item.setChecked(!item.isChecked());
    } else if (item.getItemId() == R.id.cancel_playback){
      SampleApplication.ampDownloadManager.cancelDownloads();
    }
    return true;
  }


  private final AmpDownloaderListener downloaderListener = new AmpDownloaderListener() {
    @Override
    public void downloadActionPaused(AmpDownload asset) {
      Log.e("DOWNLOADER TEST", "downloader event: PAUSED");
    }

    @Override
    public void downloadActionResumed(AmpDownload asset) {
      Log.e("DOWNLOADER TEST", "downloader event: RESUMED");
    }

    @Override
    public void downloadActionCompleted(AmpDownload asset) {
      sampleAdapter.notifyDataSetChanged();
      Log.e("DOWNLOADER TEST", "downloader event: COMPLETED");
    }

    @Override
    public void downloadActionStarted(AmpDownload asset) {
      Log.e("DOWNLOADER TEST", "downloader event: STARTED");
    }

    @Override
    public void downloadActionCancelled(AmpDownload asset) {
      Log.e("DOWNLOADER TEST", "downloader event: CANCELLED");
    }

    @Override
    public void downloadActionDeleted(AmpDownload asset) {
      sampleAdapter.notifyDataSetChanged();
      Log.e("DOWNLOADER TEST", "downloader event: DELETED");
    }

    @Override
    public void downloadActionQueued(AmpDownload asset) {
      Log.e("DOWNLOADER TEST", "downloader event: QUEUED");
    }

    @Override
    public void downloadActionFailed(AmpDownload asset, Exception exception, String message) {
      Log.e("DOWNLOADER TEST", "downloader event: FAILED");
    }
  };

  @Override
  public void onStart() {
    super.onStart();
    SampleApplication.ampDownloadManager.addListener(downloaderListener);
    sampleAdapter.notifyDataSetChanged();
  }

  @Override
  public void onStop() {
    SampleApplication.ampDownloadManager.removeListener(downloaderListener);
    super.onStop();
  }

  private void onSampleGroups(final List<SampleGroup> groups, boolean sawError) {
    if (sawError) {
      Toast.makeText(getApplicationContext(), R.string.sample_list_load_error, Toast.LENGTH_LONG)
              .show();
    }
    sampleAdapter.setSampleGroups(groups);
  }

  @Override
  public boolean onChildClick(
          ExpandableListView parent, View view, int groupPosition, int childPosition, long id) {
    UriSample sample = (UriSample) view.getTag();
    startActivity(
            sample.buildIntent(
                    /* context= */ this,
                    isNonNullAndChecked(preferExtensionDecodersMenuItem), sample.extension,
                    isNonNullAndChecked(randomAbrMenuItem)
                            ? ExoWrapper.ABR_ALGORITHM_RANDOM
                            : ExoWrapper.ABR_ALGORITHM_DEFAULT));
    return true;
  }

  private void onSampleDownloadButtonClicked(Sample sample) {
    loadingFragment.setVisibility(View.VISIBLE);
    int downloadUnsupportedStringId = getDownloadUnsupportedStringId(sample);
    if (downloadUnsupportedStringId != 0) {
      Toast.makeText(getApplicationContext(), downloadUnsupportedStringId, Toast.LENGTH_LONG)
              .show();
    } else {
      UriSample uriSample = (UriSample) sample;
      SampleApplication.ampDownloadManager.prepare(uriSample.uri, uriSample.extension);
      if(SampleApplication.ampDownloadManager.isVideoDownloaded(uriSample.uri)){
        SampleApplication.ampDownloadManager.deleteDownload(uriSample.uri, uriSample.name);
        loadingFragment.setVisibility(View.INVISIBLE);
      } else{
        StartDownloadDialogHelper dialogHelper =
                new StartDownloadDialogHelper(this, uriSample.name, uriSample.uri);
        SampleApplication.ampDownloadManager.toggleDownload(new StreamTracksListener() {
          @Override
          public void tracksLoadSuccess(List<TrackKey> trackKeys, List<String> trackTitles) {
            dialogHelper.buildDialog(trackKeys, trackTitles);
            loadingFragment.setVisibility(View.INVISIBLE);
          }

          @Override
          public void tracksLoadFail(IOException e) {
            dialogHelper.onError(e);
            loadingFragment.setVisibility(View.INVISIBLE);
          }
        });
      }
    }
  }

  private int getDownloadUnsupportedStringId(Sample sample) {
    if (sample instanceof PlaylistSample) {
      return R.string.download_playlist_unsupported;
    }
    UriSample uriSample = (UriSample) sample;
    if (uriSample.drmInfo != null) {
      return R.string.download_drm_unsupported;
    }
    if (uriSample.adTagUri != null) {
      return R.string.download_ads_unsupported;
    }
    String scheme = uriSample.uri.getScheme();
    if (!("http".equals(scheme) || "https".equals(scheme))) {
      return R.string.download_scheme_unsupported;
    }
    return 0;
  }

  private static boolean isNonNullAndChecked(@Nullable MenuItem menuItem) {
    // Temporary workaround for layouts that do not inflate the options menu.
    return menuItem != null && menuItem.isChecked();
  }

  private final class SampleListLoader extends AsyncTask<String, Void, List<SampleGroup>> {

    private boolean sawError;

    @Override
    protected List<SampleGroup> doInBackground(String... uris) {
      List<SampleGroup> result = new ArrayList<>();
      Context context = getApplicationContext();
      String userAgent = Util.getUserAgent(context, "ExoPlayerDemo");
      DataSource dataSource =
              new DefaultDataSource(context, userAgent, /* allowCrossProtocolRedirects= */ false);
      for (String uri : uris) {
        DataSpec dataSpec = new DataSpec(Uri.parse(uri));
        InputStream inputStream = new DataSourceInputStream(dataSource, dataSpec);
        try {
          readSampleGroups(new JsonReader(new InputStreamReader(inputStream, "UTF-8")), result);
        } catch (Exception e) {
          Log.e(TAG, "Error loading sample list: " + uri, e);
          sawError = true;
        } finally {
          Util.closeQuietly(dataSource);
        }
      }
      return result;
    }

    @Override
    protected void onPostExecute(List<SampleGroup> result) {
      onSampleGroups(result, sawError);
    }

    private void readSampleGroups(JsonReader reader, List<SampleGroup> groups) throws IOException {
      reader.beginArray();
      while (reader.hasNext()) {
        readSampleGroup(reader, groups);
      }
      reader.endArray();
    }

    private void readSampleGroup(JsonReader reader, List<SampleGroup> groups) throws IOException {
      String groupName = "";
      ArrayList<Sample> samples = new ArrayList<>();

      reader.beginObject();
      while (reader.hasNext()) {
        String name = reader.nextName();
        switch (name) {
          case "name":
            groupName = reader.nextString();
            break;
          case "samples":
            reader.beginArray();
            while (reader.hasNext()) {
              samples.add(readEntry(reader, false));
            }
            reader.endArray();
            break;
          case "_comment":
            reader.nextString(); // Ignore.
            break;
          default:
            throw new ParserException("Unsupported name: " + name);
        }
      }
      reader.endObject();

      SampleGroup group = getGroup(groupName, groups);
      group.samples.addAll(samples);
    }

    private Sample readEntry(JsonReader reader, boolean insidePlaylist) throws IOException {
      String sampleName = null;
      Uri uri = null;
      String extension = null;
      String drmScheme = null;
      String drmLicenseUrl = null;
      String[] drmKeyRequestProperties = null;
      boolean drmMultiSession = false;
      ArrayList<UriSample> playlistSamples = null;
      String adTagUri = null;
      String sphericalStereoMode = null;

      reader.beginObject();
      while (reader.hasNext()) {
        String name = reader.nextName();
        switch (name) {
          case "name":
            sampleName = reader.nextString();
            break;
          case "uri":
            uri = Uri.parse(reader.nextString());
            break;
          case "extension":
            extension = reader.nextString();
            break;
          case "drm_scheme":
            Assertions.checkState(!insidePlaylist, "Invalid attribute on nested item: drm_scheme");
            drmScheme = reader.nextString();
            break;
          case "drm_license_url":
            Assertions.checkState(!insidePlaylist,
                    "Invalid attribute on nested item: drm_license_url");
            drmLicenseUrl = reader.nextString();
            break;
          case "drm_key_request_properties":
            Assertions.checkState(!insidePlaylist,
                    "Invalid attribute on nested item: drm_key_request_properties");
            ArrayList<String> drmKeyRequestPropertiesList = new ArrayList<>();
            reader.beginObject();
            while (reader.hasNext()) {
              drmKeyRequestPropertiesList.add(reader.nextName());
              drmKeyRequestPropertiesList.add(reader.nextString());
            }
            reader.endObject();
            drmKeyRequestProperties = drmKeyRequestPropertiesList.toArray(new String[0]);
            break;
          case "drm_multi_session":
            drmMultiSession = reader.nextBoolean();
            break;
          case "playlist":
            Assertions.checkState(!insidePlaylist, "Invalid nesting of playlists");
            playlistSamples = new ArrayList<>();
            reader.beginArray();
            while (reader.hasNext()) {
              playlistSamples.add((UriSample) readEntry(reader, true));
            }
            reader.endArray();
            break;
          case "ad_tag_uri":
            adTagUri = reader.nextString();
            break;
          case "spherical_stereo_mode":
            Assertions.checkState(
                    !insidePlaylist, "Invalid attribute on nested item: spherical_stereo_mode");
            sphericalStereoMode = reader.nextString();
            break;
          default:
            throw new ParserException("Unsupported attribute name: " + name);
        }
      }
      reader.endObject();
      DrmInfo drmInfo =
              drmScheme == null
                      ? null
                      : new DrmInfo(drmScheme, drmLicenseUrl, drmKeyRequestProperties, drmMultiSession);
      if (playlistSamples != null) {
        UriSample[] playlistSamplesArray = playlistSamples.toArray(
                new UriSample[playlistSamples.size()]);
        return new PlaylistSample(sampleName, drmInfo, playlistSamplesArray);
      } else {
        return new UriSample(
                sampleName,
                drmInfo,
                uri,
                extension,
                adTagUri,
                sphericalStereoMode);
      }
    }

    private SampleGroup getGroup(String groupName, List<SampleGroup> groups) {
      for (int i = 0; i < groups.size(); i++) {
        if (Util.areEqual(groupName, groups.get(i).title)) {
          return groups.get(i);
        }
      }
      SampleGroup group = new SampleGroup(groupName);
      groups.add(group);
      return group;
    }

  }

  private final class SampleAdapter extends BaseExpandableListAdapter implements OnClickListener {

    private List<SampleGroup> sampleGroups;

    public SampleAdapter() {
      sampleGroups = Collections.emptyList();
    }

    public void setSampleGroups(List<SampleGroup> sampleGroups) {
      this.sampleGroups = sampleGroups;
      notifyDataSetChanged();
    }

    @Override
    public Sample getChild(int groupPosition, int childPosition) {
      return getGroup(groupPosition).samples.get(childPosition);
    }

    @Override
    public long getChildId(int groupPosition, int childPosition) {
      return childPosition;
    }

    @Override
    public View getChildView(int groupPosition, int childPosition, boolean isLastChild,
                             View convertView, ViewGroup parent) {
      View view = convertView;
      if (view == null) {
        view = getLayoutInflater().inflate(R.layout.sample_list_item, parent, false);
        View downloadButton = view.findViewById(R.id.download_button);
        downloadButton.setOnClickListener(this);
        downloadButton.setFocusable(false);
      }
      initializeChildView(view, getChild(groupPosition, childPosition));
      return view;
    }

    @Override
    public int getChildrenCount(int groupPosition) {
      return getGroup(groupPosition).samples.size();
    }

    @Override
    public SampleGroup getGroup(int groupPosition) {
      return sampleGroups.get(groupPosition);
    }

    @Override
    public long getGroupId(int groupPosition) {
      return groupPosition;
    }

    @Override
    public View getGroupView(int groupPosition, boolean isExpanded, View convertView,
                             ViewGroup parent) {
      View view = convertView;
      if (view == null) {
        view =
                getLayoutInflater()
                        .inflate(android.R.layout.simple_expandable_list_item_1, parent, false);
      }
      ((TextView) view).setText(getGroup(groupPosition).title);
      return view;
    }

    @Override
    public int getGroupCount() {
      return sampleGroups.size();
    }

    @Override
    public boolean hasStableIds() {
      return false;
    }

    @Override
    public boolean isChildSelectable(int groupPosition, int childPosition) {
      return true;
    }

    @Override
    public void onClick(View view) {
      if(!onClickBlocked) {
        blockAllButtons();
        onSampleDownloadButtonClicked((Sample) view.getTag());
      }
    }

    boolean onClickBlocked = false;

    private void blockAllButtons(){
      onClickBlocked = true;
      new Handler().postDelayed(() -> onClickBlocked = false, 1000);
    }

    private void initializeChildView(View view, Sample sample) {
      view.setTag(sample);
      TextView sampleTitle = view.findViewById(R.id.sample_title);
      sampleTitle.setText(sample.name);

      boolean canDownload = getDownloadUnsupportedStringId(sample) == 0;
      boolean isDownloaded = canDownload && isVideoDownloaded(((UriSample) sample).uri);
      ImageButton downloadButton = view.findViewById(R.id.download_button);
      downloadButton.setTag(sample);
      downloadButton.setColorFilter(
              canDownload ? (isDownloaded ? 0xFF42A5F5 : 0xFFBDBDBD) : 0xFFEEEEEE);
      downloadButton.setImageResource(
              isDownloaded ? R.drawable.ic_download_done : R.drawable.ic_download);
    }

    private boolean isVideoDownloaded(Uri uri) {
      AmpDownloadManager downloadManager = SampleApplication.ampDownloadManager;
      return downloadManager.isVideoDownloaded(uri) || downloadManager.isVideoDownloaded(SsUtil.fixManifestUri(uri));
    }
  }

  private static final class SampleGroup {

    public final String title;
    public final List<Sample> samples;

    public SampleGroup(String title) {
      this.title = title;
      this.samples = new ArrayList<>();
    }

  }

  private static final class DrmInfo {
    public final String drmScheme;
    public final String drmLicenseUrl;
    public final String[] drmKeyRequestProperties;
    public final boolean drmMultiSession;

    public DrmInfo(
            String drmScheme,
            String drmLicenseUrl,
            String[] drmKeyRequestProperties,
            boolean drmMultiSession) {
      this.drmScheme = drmScheme;
      this.drmLicenseUrl = drmLicenseUrl;
      this.drmKeyRequestProperties = drmKeyRequestProperties;
      this.drmMultiSession = drmMultiSession;
    }

    public void updateIntent(Intent intent) {
      Assertions.checkNotNull(intent);
      intent.putExtra(ExoWrapper.DRM_SCHEME_EXTRA, drmScheme);
      intent.putExtra(ExoWrapper.DRM_LICENSE_URL_EXTRA, drmLicenseUrl);
      intent.putExtra(ExoWrapper.DRM_KEY_REQUEST_PROPERTIES_EXTRA, drmKeyRequestProperties);
      intent.putExtra(ExoWrapper.DRM_MULTI_SESSION_EXTRA, drmMultiSession);
    }
  }

  private abstract static class Sample {
    public final String name;
    public final DrmInfo drmInfo;

    public Sample(String name, DrmInfo drmInfo) {
      this.name = name;
      this.drmInfo = drmInfo;
    }

    public Intent buildIntent(
            Context context, boolean preferExtensionDecoders, String abrAlgorithm) {
      Intent intent = new Intent(context, PlayerActivity.class);
      intent.putExtra(ExoWrapper.PREFER_EXTENSION_DECODERS_EXTRA, preferExtensionDecoders);
      intent.putExtra(ExoWrapper.ABR_ALGORITHM_EXTRA, abrAlgorithm);
      if (drmInfo != null) {
        drmInfo.updateIntent(intent);
      }
      return intent;
    }

  }

  private static final class UriSample extends Sample {

    public final Uri uri;
    public final String extension;
    public final String adTagUri;
    public final String sphericalStereoMode;

    public UriSample(
            String name,
            DrmInfo drmInfo,
            Uri uri,
            String extension,
            String adTagUri,
            String sphericalStereoMode) {
      super(name, drmInfo);
      this.uri = uri;
      this.extension = extension;
      this.adTagUri = adTagUri;
      this.sphericalStereoMode = sphericalStereoMode;
    }

    public Intent buildIntent(
            Context context, boolean preferExtensionDecoders, String extension, String abrAlgorithm) {
      Intent intent =  super.buildIntent(context, preferExtensionDecoders, abrAlgorithm);
      intent.setData(uri)
              .putExtra(ExoWrapper.EXTENSION_EXTRA, extension)
              .putExtra(ExoWrapper.AD_TAG_URI_EXTRA, adTagUri)
              .putExtra(ExoWrapper.SPHERICAL_STEREO_MODE_EXTRA, sphericalStereoMode)
              .setAction(ExoWrapper.ACTION_VIEW);
      return intent;

    }

  }

  private static final class PlaylistSample extends Sample {

    public final UriSample[] children;

    public PlaylistSample(
            String name,
            DrmInfo drmInfo,
            UriSample... children) {
      super(name, drmInfo);
      this.children = children;
    }

    @Override
    public Intent buildIntent(
            Context context, boolean preferExtensionDecoders, String abrAlgorithm) {
      String[] uris = new String[children.length];
      String[] extensions = new String[children.length];
      for (int i = 0; i < children.length; i++) {
        uris[i] = children[i].uri.toString();
        extensions[i] = children[i].extension;
      }
      return super.buildIntent(context, preferExtensionDecoders, abrAlgorithm)
              .putExtra(ExoWrapper.URI_LIST_EXTRA, uris)
              .putExtra(ExoWrapper.EXTENSION_LIST_EXTRA, extensions)
              .setAction(ExoWrapper.ACTION_VIEW_LIST);
    }

  }

}
